/*----------------------------------------------------------------------*/
/*      sortl.c         list sort                                       */
/*                                                                      */
/*      Jeff Reid       13AUG14  16:00                                  */
/*----------------------------------------------------------------------*/
#define _CRT_SECURE_NO_WARNINGS 1
#include <stdio.h>
#include <stdlib.h>

#define NUMNODES (1024*4096)            /* number of nodes */
#define NUMLISTS 32                     /* number of lists */

typedef unsigned long long UI64;

/*      MergeLists() casts an instance of LIST to NODE      */
/*      and uses NODE.next as the equivalent of LIST.first  */

typedef struct _NODE{
struct _NODE * next;        /* must be first member of NODE */
UI64 data;
}NODE, *PNODE;

typedef struct _LIST{
NODE *first;                /* must be first member of LIST */
}LIST, *PLIST;

/*----------------------------------------------------------------------*/
/*      data                                                            */
/*----------------------------------------------------------------------*/
static LIST aList[NUMLISTS];    /* array of lists */

static FILE * fpDst;            /* dst (output) file */

/*----------------------------------------------------------------------*/
/*      code                                                            */
/*----------------------------------------------------------------------*/
void SortList(LIST *pList);
void MergeNode(NODE *pNode);
void MergeLists(LIST *plDst, LIST *plSrc1, LIST *plSrc2);
NODE * GetFirstNode(LIST * pSrc);
LIST * CreateList(LIST * pList, int count);

/*----------------------------------------------------------------------*/
/*      main                                                            */
/*----------------------------------------------------------------------*/
int main(void)
{
LIST dList;
NODE *pNode;
void *pMem;

    if(CreateList(&dList, NUMNODES) == NULL)    /* create list */
        return(0);
    pMem = dList.first;                         /* save malloc ptr */
    SortList(&dList);                           /* sort the list */

    fpDst = fopen("dstl.bin", "wb");            /* write data */
    if(fpDst == NULL){
        goto exit0;}
    while(NULL != (pNode = GetFirstNode(&dList)))
        fwrite(&(pNode->data), 1, sizeof(UI64), fpDst);
    fclose(fpDst);

exit0:
    free(pMem);                                 /* free memory */
    return(0);
}
 
/*----------------------------------------------------------------------*/
/*      SortList                                                        */
/*----------------------------------------------------------------------*/
void SortList(LIST *pList)
{
NODE * pNode;
LIST mList;                             /* merged list */
int i;

    if(pList == NULL)                   /* check for null ptr */
        return;

    /* merge nodes into aList[] */
    while(NULL != (pNode = GetFirstNode(pList))){
        MergeNode(pNode);
    }

    /* find 1st non empty aList[] */
    for(i = 0; (i < NUMLISTS) && (aList[i].first == NULL); i++);
    if(i >= NUMLISTS){                  /* if all empty */
        pList->first = NULL;            /*   return null list */
        return;
    }

    pList->first = aList[i].first;      /* plist = aList[i] */
    aList[i].first = NULL;              /* clear aList[i] (optional) */

    for(i++; i < NUMLISTS; i++){        /* merge remaining aList[] */
        if(aList[i].first != NULL){
            MergeLists(&mList, &aList[i], pList);
            pList->first = mList.first;
        }
    }
}

/*----------------------------------------------------------------------*/
/*      MergeNode   merge a node into the array of lists                */
/*----------------------------------------------------------------------*/
void MergeNode(NODE *pNode)
{
LIST sList;                             /* source list */
LIST mList;                             /* merged list */
int i;

    if(pNode == NULL)                   /* return if null ptr */
        return;

    sList.first = pNode;                /* src list = node */

    /* merge into array */
    for(i = 0; (i < NUMLISTS) && (aList[i].first != NULL); i++){
        MergeLists(&mList, &aList[i], &sList);
        sList.first = mList.first;
    }
    if(i == NUMLISTS)                   /* update array */
        i--;
    aList[i].first = sList.first;
}

/*----------------------------------------------------------------------*/
/*      MergeLists  dst = merge(src1, src2)                             */
/*----------------------------------------------------------------------*/
void MergeLists(LIST *plDst, LIST *plSrc1, LIST *plSrc2)
{
NODE *pDst, *pSrc1, *pSrc2;             /* node ptrs */

    pDst = (NODE *)plDst;               /* treat list same as node */
    pSrc1 = GetFirstNode(plSrc1);       /* init ptrs to nodes */
    pSrc2 = GetFirstNode(plSrc2);

    while(1){
        if(pSrc2->data < pSrc1->data){  /* if src2 < src1 */
            pDst->next = pSrc2;
            pDst = pDst->next;
            /* if end of src2, dst = rest of src1 */
            if(NULL == (pSrc2 = GetFirstNode(plSrc2))){
                pSrc1->next = plSrc1->first;
                pDst->next = pSrc1;
                plSrc1->first = NULL;
                break;
            }
        } else {                        /* src1 <= src2 */
            pDst->next = pSrc1;
            pDst = pDst->next;
            /* if end of src1, dst = rest of src2 */
            if(NULL == (pSrc1 = GetFirstNode(plSrc1))){
                pSrc2->next = plSrc2->first;
                pDst->next = pSrc2;
                plSrc2->first = NULL;
                break;
            }
        }
    }
}

/*----------------------------------------------------------------------*/
/*      GetFirstNode                                                    */
/*----------------------------------------------------------------------*/
NODE * GetFirstNode(LIST * pSrc)
{
NODE * pNode;

    if(pSrc == NULL)                        /* return if null ptr */
        return NULL;
    pNode = pSrc->first;                    /* get node */
    if(pNode != NULL){
        pSrc->first = pNode->next;          /* advance list ptr */
        pNode->next = NULL;                 /* reset node ptr */
    }
    return pNode;
}

/*----------------------------------------------------------------------*/
/*      create list     used for test                                   */
/*----------------------------------------------------------------------*/
LIST * CreateList(LIST * pList, int count)
{
NODE *pNode;
int i;
UI64 r;

    if(pList == NULL)                   /* check for null ptr */
        return NULL;
    /* allocate space for nodes */
    pList->first = malloc(count * sizeof(NODE));
    if(pList->first == NULL)
        return NULL;
    pNode = pList->first;               /* init nodes */
    for(i = 0; i < count; i++){
        r  = (((UI64)((rand()>>4) & 0xff))<< 0);
        r += (((UI64)((rand()>>4) & 0xff))<< 8);
        r += (((UI64)((rand()>>4) & 0xff))<<16);
        r += (((UI64)((rand()>>4) & 0xff))<<24);
        r += (((UI64)((rand()>>4) & 0xff))<<32);
        r += (((UI64)((rand()>>4) & 0xff))<<40);
        r += (((UI64)((rand()>>4) & 0xff))<<48);
        r += (((UI64)((rand()>>4) & 0xff))<<56);
        pNode->data = r;
        pNode->next = pNode+1;
        pNode++;
    }
    (--pNode)->next = NULL;
    return pList;
}
